import atexit
from Queue import Queue, Empty as QueueEmptyError
from threading import Thread
from .singleton import Singleton


class Worker(Thread):

    '''Thread executing tasks from a given tasks queue'''

    def __init__(self, tasks):
        Thread.__init__(self)
        self.tasks = tasks
        self.daemon = True
        self.start()

    def run(self):
        while True:
            try:
                cls, func_name, args, kwargs = self.tasks.get_nowait()
                getattr(cls, func_name)(*args, **kwargs)
                self.tasks.task_done()
            except QueueEmptyError:
                break

class ThreadPool:
    __metaclass__ = Singleton
    '''Pool of threads consuming tasks from a queue'''

    def __init__(self, max_threads):
        self.tasks = Queue(max_threads * 100)
        self._max_threads = max_threads
        self._workers = []
        self._active = True

    def _clean_finished(self):
        self._workers = filter(lambda w: w.is_alive(), self._workers)

    def add_task(self, cls, func_name, *args, **kwargs):
        if not self._active:
            raise RuntimeError('ThreadPool is not active')

        self._clean_finished()
        self.tasks.put((cls, func_name, args, kwargs))

        if self.tasks.qsize() > len(self._workers) < self._max_threads:
            self._workers.append(Worker(self.tasks))

    def shutdown(self):
        self._active = False
        self.tasks.join()

# FIXME: Configure this value from outside
tpool = ThreadPool(10)
atexit.register(tpool.shutdown)
